%{
===========================================================================
Code provided by Yichuan (Charlie) Tang
Permission is granted for anyone to copy, use, modify, or distribute this
program and accompanying programs and documents for any purpose, provided
this copyright notice is retained and prominently displayed, along with
a note saying that the original programs are available from our
web page.
The programs and documents are distributed without any warranty, express or
implied.  As the programs were written for research purposes only, they 
have not been tested to the degree that would be advisable in any important
application.  All use of these programs is entirely at the user's own risk.
===========================================================================

%CT 3/2012 - learning a mixture of tensor analyzers
           - this supports more than 2 sets of latent factors

% global variables:
% xxTrain - global data matrix:
%           data matrix N x D x batches
% xxValid - N by D validation data (for early stopping)

input:
params.d1  - z1 latent dimensionality
params.d2  - z2 latent dimensionality
params.C - number of mixture components
params.iso - 1 if noise Psi should be isotropic/sphereical
params.eq - 1 if all components should share the same Psi(diagonal covariance)
params.GPU - use GPU or not
params.epsilon - relative increase for convergence %default 1e-3
params.max_iters - limit number of iterations

params.PreWts.Psi - must be visible dim (D) by C. C is the number of mixtures
params.PreWts.M - must be D by C - these are the means
params.PreWts.W - D by (d1+d2) by C - 2-biases
params.PreWts.T - D by (d1*d2) by C  matricized Tensor

output:
logl_epoch  : row vector of log-likelihoods after each EM iteration
MoTA  : structure containing the Mixture of Factor Analyzers
    T [D x (d1*d2) x C]     - tensor (matricized)     
    W [D x (d1+d2) x C]     - 2-biases    
    M [D x C]               - d-dimensional means    
    Psi [D x C]             - diagonal noise levels    
    mix [C x 1 ]            - mixing proportions
%}

function [ logl_epoch, MoTA ] = mta3( params )

global xxTrain;
global xxValid;
global labels;

if isfield( params, 'epsilon')
    epsilon = params.epsilon;
else
    epsilon     = 1e-3;  % relative log-likelihood increase threshold for convergence
end

if isfield( params, 'max_iters')
    max_iters = params.max_iters;
else
    max_iters = 1000; % maximum number of EM iterations
end
min_iters  = 5;       % minimum number of EM iterations

[N, D, nBatches] = size(xxTrain);

if 1
    fprintf('\n--> MATLAB EM for MoTA3 \n');
	fprintf('    mixture components: %d\n', params.C);
	fprintf('    d: %d \n', params.d);
	fprintf('    data: %d points in %d dimensions batches:%d\n', N, D, nBatches);
    fprintf('    Learning nGibbs: %d   nLLSamples: %d\n', params.nGibbsSamples, params.nLLSamples);
        	
    if params.iso==1
        fprintf(' !!! Using isotropic noise model\n');
    else
        fprintf('Non-Isotropic noise model\n');
    end
    if params.eq==1
        fprintf(' !!! Equal noise model in components\n');
    else
        fprintf('Non-equal noise model\n');
    end
end

MoTA =[];
% ----------- Initialization ---------------
if isfield( params.PreWts, 'M')
    MoTA.M = double(params.PreWts.M); 
else
    assert(false);
    %MoFA.M     = randperm(N); 
    %MoFA.M     = X(:,MoFA.M(1:C));    % means
end

if  isfield( params.PreWts, 'Psi')
    MoTA.Psi = double(params.PreWts.Psi);
else
    assert(false);
    %MoFA.Psi = 100*var(X(:))*ones(D,C);              % noise variance
end

if isfield( params.PreWts, 'W')
    MoTA.W = double(params.PreWts.W);
else
    assert(false);
    %MoFA.W = 0.01*randn(D,d,C);                      % factor loadings
end

if isfield( params.PreWts, 'mix')
    MoTA.mix   = params.PreWts.mix;                 % mixing weights
else     
    MoTA.mix   = ones(params.C,1)/params.C;  
end

if isfield( params.PreWts, 'T')
    MoTA.T   = params.PreWts.T;                 % mixing weights
else     
    assert(false);
end

MoTA.d = params.d;
assert( min(MoTA.d) >= 2);
%a bug will be caused in TensorMatrixProd_zs if a d==1

% ----------- end ---------------

if params.C==1
    assert( all( size(MoTA.T) == [D prod(MoTA.d) ]));
    assert( all( size(MoTA.W) == [D sum(MoTA.d) ]));
else
    assert( all( size(MoTA.T) == [D prod(MoTA.d) params.C]));
    assert( all( size(MoTA.W) == [D sum(MoTA.d) params.C]));
end
assert( all( size(MoTA.M) == [D params.C]));
assert( all( size(MoTA.Psi) == [D params.C]));
assert( all( size(MoTA.mix) == [params.C 1]));

if exist('xxValid', 'var') && ~isempty(xxValid)
    likv = eval_mta3(double(xxValid)', MoTA, params.nLLSamples);
    valid_logl = mean(likv);
    fprintf('\n*** Initial xxValid:%f ***\n', valid_logl);
else
    valid_logl = -inf;
end

if ~isempty(labels)
    fprintf('\n****Labels are available!');
end

logl_epoch=zeros(1,max_iters);

numbatchesUse = min(nBatches, params.num_mbs);
inds = randperm( nBatches );
inds = inds(1:numbatchesUse);

for bb = 1:length(inds) %loop thru minibatches
    
    for iter=1:max_iters; % EM loops
        
        mstep_p =[];
        mstep_p.eq = params.eq;
        mstep_p.iso = params.iso;
        mstep_p.GPU = params.GPU;
        mstep_p.QTHRESH = params.QTHRESH;
        mstep_p.MIN_PSI = params.MIN_PSI;
        
        X = xxTrain(:,:,inds(bb));
        
        fprintf('\n mta3: EM minibatch:%d/%d iter:%d/%d \t', bb, length(inds), iter, max_iters);
        fprintf('Estep, ');
        [Q, LL, zs] = mta3_E_step( double(X)', MoTA,  ...
                params.nGibbsSamples, params.nLLSamples, labels);% E-step        
        LogL(iter) = sum(LL);        
                       
        fprintf(' Mstep ');
        MoTA_new = mta3_M_step( mstep_p, double(X)', Q, zs, MoTA, ...
                params.nUpdateIteratons, params.ZM, params.learnrate, params.wtcost);% M-step
        fprintf('.');        
        if iter > 1
            rel_change = (LogL(iter)-LogL(iter-1)) / abs(mean(LogL(iter-1:iter)));
            
            if rel_change < 0
                fprintf('--> LogL decreased in iter: %d,  rel. change % .5f ;',iter,rel_change);
                %break;
            end
            
            fprintf(' iter %3d minib:[%d]%d/%d avgLL(or mse):%.7f rel. inc. % .5f  ',iter, inds(bb), bb, numbatchesUse, LogL(iter)/N, rel_change);

            %if dia>1; clf;set(gcf,'Double','on');imagesc(Q);xlabel('data point index');ylabel('mixture component index');set(gca,'ytick',1:C);colormap gray;colorbar;title('Posterior distributions on mixture components given data point');drawnow;end
            if (rel_change < epsilon) && (iter > min_iters);
               % break;
            end

        elseif iter == 1
            MoTA = MoTA_new;
            fprintf(' iteration %3d  [%d]%d/%d avgLL(or mse):%.2f Logl % .2f ',iter, inds(bb), bb, numbatchesUse, LogL(iter)/N, LogL(iter));
        end
        
                
        if exist('xxValid', 'var') && ~isempty(xxValid)
            likv = eval_mta3(double(xxValid)', MoTA_new, params.nLLSamples);
            valid_logl_new = mean(likv);
            fprintf(' xxValid:% .2f ', valid_logl_new);
            
            if valid_logl_new < valid_logl && params.EarlyStop == 1
                fprintf('  --> Early Stopping! %d ;', iter );
                break;
            end
            valid_logl = valid_logl_new;
            
            logl_epoch(iter) = valid_logl;
        end
        
        MoTA = MoTA_new;
    end
end
